/*******************************************************************************
 * SVMlin
 * Copyright (c) 2006 Vikas Sindhwani at the University of Chicago.
 * Adapted to Percolator by Lukas Käll at the University of Washington
 *
 *******************************************************************************/
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <cmath>
using namespace std;
#include "Globals.h"
#include "sslF.h"

#include <stdarg.h>
#include <omp.h>

extern "C" {
  extern double dnrm2_(int *, double *, int *);
  extern double ddot_(int *, double *, int *, double *, int *);
  extern int daxpy_(int *, double *, double *, int *, double *, int *);
  extern int dscal_(int *, double *, double *, int *);
  extern int dgemv_(char *, int *, int *,
		    double *, double *, int *,
		    double *, int *,  double *,
		    double *, int *);
}

Reduce_Vectors::Reduce_Vectors(int size)
{
  // #ifdef NOMAP
  nr_thread = omp_get_max_threads();
  // #else
  // nr_thread = 3 * THREADS;
  // #endif
  this->size = size;
  this->size0 = size-1;
  tmp_array = new double*[nr_thread];
  for(int i = 0; i < nr_thread; i++)
    tmp_array[i] = new double[size];
}

Reduce_Vectors::~Reduce_Vectors(void)
{
  for(int i = 0; i < nr_thread; i++)
    delete[] tmp_array[i];
  delete[] tmp_array;
}

void Reduce_Vectors::init(void)
{
#pragma omp parallel for schedule(static) // num_threads(nr_thread)
  for(int i = 0; i < size; i++)
    for(int j = 0; j < nr_thread; j++)
      tmp_array[j][i] = 0.0;
}

void Reduce_Vectors::sum_scale_x(double scalar, double* x)
{
  int thread_id = omp_get_thread_num();
  int inc = 1;

  daxpy_(&size, &scalar, x, &inc, tmp_array[thread_id], &inc);
  // daxpy_(&size0, &scalar, x, &inc, tmp_array[thread_id], &inc);
  // tmp_array[thread_id][size-1] += scalar;
}

void Reduce_Vectors::reduce_sum(double* v)
{
#pragma omp parallel for schedule(static) // num_threads(nr_thread)
  for(int i = 0; i < size; i++)
    {
      v[i] = 0;
      for(int j = 0; j < nr_thread; j++)
	v[i] += tmp_array[j][i];
    }
}

void Reduce_Vectors::reduce_sum_cgls(double* v)
{
#pragma omp parallel for schedule(static) // num_threads(nr_thread)
  for(int i = 0; i < size; i++)
    {
      for(int j = 0; j < nr_thread; j++)
	v[i] += tmp_array[j][i];
    }
}

double fun_omp(double *w, int w_size, int l,
	       const double* y, double* z, const double* C)
{
  unsigned int i;
  int inc = 1;
  double f=0;

// f = ddot_(&w_size, w, &inc, w, &inc) / 2.0;
#pragma omp parallel for private(i) reduction(+:f) schedule(guided) // num_threads(THREADS)
 for(i=0;i<l;i++)
   {
     // z[i]=y[i]*(ddot_(&w_size, w, &inc, X[i], &inc) + w[w_size - 1]);
     z[i]*=y[i]; //*(ddot_(&w_size, w, &inc, X[i], &inc) + w[w_size - 1]);
     double d = 1-z[i];
     if (d > 0)
       f += C[i]*d*d;
   }
 
 return(f);
}

#ifdef GPUGRAD
void subXTv_omp(double *v, const double* y, const double* C,
		double *XTv, int w_size,
		double* X, const int* Id, int sizeI, Reduce_Vectors *reduce_vectors)
{
  int i;
  int inc = 1;
  double a = 0;

  reduce_vectors->init();

#pragma omp parallel for private(i) schedule(guided) // num_threads(THREADS)
  for(i=0;i<sizeI;i++){
    int h = Id[i];
    reduce_vectors->sum_scale_x(C[h]*y[h]*(v[h]-1), X + h*w_size);
  }
  
  reduce_vectors->reduce_sum(XTv);
}
#else
void subXTv_omp(double *v, double *XTv, int w_size,
		double* X, const int* Id, int sizeI, Reduce_Vectors *reduce_vectors)
{
  int i;
  int inc = 1;
  double a = 0;

  reduce_vectors->init();

#pragma omp parallel for private(i) schedule(guided)
  for(i=0;i<sizeI;i++)
    reduce_vectors->sum_scale_x(v[i], X + Id[i] * w_size);
  
  reduce_vectors->reduce_sum(XTv);
}
#endif

#ifndef GPUGRAD
int grad_omp(double *w, double *g, int w_size, int l,
	     const double* y, double* z, const double* C, double* X,
	     int* Id, Reduce_Vectors *reduce_vectors)
{
  int i;
  int sizeI = 0;
  for (i=0;i<l;i++){
    if (z[i] < 1) {
      z[sizeI] = C[i]*y[i]*(z[i]-1);
      Id[sizeI] = i;
      sizeI++;
    }
  }
  subXTv_omp(z, g, w_size, X, Id, sizeI, reduce_vectors);

  for(i=0;i<w_size;i++)
    g[i] = w[i] + 2*g[i];
  return(sizeI);
}
#else
void grad_omp(double *w, double *g, int w_size, int l,
	     const double* y, double* z, const double* C, double* X,
	      int* Id, Reduce_Vectors *reduce_vectors, int sizeI)
{
  int i;
  subXTv_omp(z, C, y,
	     g, w_size, X, Id, sizeI, reduce_vectors);
  for(i=0;i<w_size;i++)
    g[i] = w[i] + 2*g[i];
}
#endif

void Hv_omp(double *s, double *Hs, int w_size,
	    const double* C, double* X, const int *Id, int sizeI,
	    Reduce_Vectors *reduce_vectors)
{
  int i;
  int inc = 1;

  reduce_vectors->init();
#pragma omp parallel for private(i) schedule(guided) // num_threads(THREADS)
  for(i=0;i<sizeI;i++)
    {
      double* xi = X + Id[i] * w_size;
      // double xTs = C[Id[i]]*(ddot_(&w_size, s, &inc, xi, &inc) + s[w_size - 1]);
      double xTs = C[Id[i]]*ddot_(&w_size, s, &inc, xi, &inc);
      reduce_vectors->sum_scale_x(xTs, xi);
    }
  reduce_vectors->reduce_sum(Hs);
  for(i=0;i<w_size;i++)
    Hs[i] = s[i] + 2*Hs[i];
}

int omp_thread_count() {
  int n = 0;
#pragma omp parallel reduction(+:n)
   n += 1;
   return n;
}
