#######################################################
# 
# Written by John Halloran <jthalloran@ucdavis.edu>
#
# Copyright (C) 2020 John Halloran
# Licensed under the Open Software License version 3.0
# See COPYING or http://opensource.org/licenses/OSL-3.0
#
#######################################################
#
# If you use any of the code below, please cite:
# J. T. Halloran and D. M. Rocke.
# GPU-Accelerated Primal Learning for Extremely Fast Large-Scale Classification,
# Advances in Neural Information Processing Systems (NeurIPS) 2020
#
#######################################################

import torch
from torch.autograd import Variable

from sklearn.datasets import load_svmlight_file
from sklearn.model_selection import train_test_split
from sklearn import preprocessing
import numpy as np
import timeit
from sklearn.linear_model import LogisticRegression as lr

class LogisticRegression(torch.nn.Module):
    def __init__(self, input_dim, output_dim):
        super(LogisticRegression, self).__init__()
        self.linear = torch.nn.Linear(input_dim, output_dim)

    def forward(self, x):
        outputs = torch.sigmoid(self.linear(x))
        return outputs

    def train(self, images, labels, epochs = 3000):
        for epoch in range(epochs):
            optimizer.zero_grad()
            outputs = model(images)
            loss = criterion(outputs, labels)
            loss.backward()
            optimizer.step()
        return loss.item()

    def train_lbfgs(self, images, labels, epochs = 100, doPrint = False):
        for epoch in range(epochs):

            def closure():
                optimizer.zero_grad()
                outputs = model(images)
                loss = criterion(outputs, labels)
                loss.backward()
                return loss
            optimizer.step(closure)

            outputs = model(images)
            loss = criterion(outputs, labels)
        return loss.item()

X,y = load_svmlight_file("rcv1_train.binary")
# Make the train-test split
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.1, random_state=42)

logit = lr(solver="liblinear")
start_time = timeit.default_timer()

logit.fit(X_train, y_train)

elapsed = timeit.default_timer() - start_time
print "Liblinear Training took %f seconds" % (elapsed)
predicted = logit.predict(X_test)
correct= (predicted == y_test).sum()
accuracy = 100. * float(correct) / float(len(y_test))
print "Liblinear Test accuracy:%f" % (accuracy)

# Convert to numpy ararys
X_train = X_train.A
X_test  = X_test.A
y_train = np.array([1 if yi > 0 else 0 for yi in y_train])
y_test = np.array([1 if yi > 0 else 0 for yi in y_test])

# Normalize the training data
preprocessing.scale(X_train, copy = False)
preprocessing.scale(X_test, copy = False)

l= X_train.shape

batch_size = len(y_train)
batches = int((len(y_train) / batch_size))
epochs = 3000
input_dim = l[1]
output_dim = 1
lr_rate = 0.001

########################################################
############################ Cuda Training
########################################################
print "Training pytorch models using GPU"
DEVICE = torch.device("cuda:0")
########################################################
############################ Gradient Descent
########################################################

# Model and training set up
model = LogisticRegression(input_dim, output_dim).to(DEVICE)
criterion = torch.nn.BCELoss().to(DEVICE)
optimizer = torch.optim.SGD(model.parameters(), lr=lr_rate)

y_train.resize(len(y_train),1)
y_test.resize(len(y_test),1)
total = len(y_train)
total_test = len(y_test)
images = Variable(torch.from_numpy(X_train)).float()
labels = Variable(torch.from_numpy(y_train)).float()
test_images = Variable(torch.from_numpy(X_test)).float()
test_labels = Variable(torch.from_numpy(y_test)).float()

images = images.to(DEVICE)
labels = labels.to(DEVICE)
test_images = test_images.to(DEVICE)
test_labels = test_labels.to(DEVICE)

print "Iterations = %d" % (epochs)

start_time = timeit.default_timer()
lo = model.train(images, labels, epochs)
elapsed = timeit.default_timer() - start_time
print "Cuda training using gradient descent took %f seconds" % (elapsed)

# Calculate test accuracy
outputs = model(test_images)
predicted = torch.round(outputs)
correct= (predicted == test_labels).sum()
accuracy = 100. * float(correct.item())/float(total_test)
print "Gradient descent: train loss = %f, test accuracy = %f" % (lo, accuracy)

# Free up device memory
torch.cuda.empty_cache()
########################################################
############################ L-BFGS
########################################################

epochs = 100
# Model and training set up
model = LogisticRegression(input_dim, output_dim).to(DEVICE)
criterion = torch.nn.BCELoss().to(DEVICE)
optimizer = torch.optim.LBFGS(model.parameters()) 

y_train.resize(len(y_train),1)
y_test.resize(len(y_test),1)
total = len(y_train)
total_test = len(y_test)
images = Variable(torch.from_numpy(X_train)).float()
labels = Variable(torch.from_numpy(y_train)).float()
test_images = Variable(torch.from_numpy(X_test)).float()
test_labels = Variable(torch.from_numpy(y_test)).float()

images = images.to(DEVICE)
labels = labels.to(DEVICE)
test_images = test_images.to(DEVICE)
test_labels = test_labels.to(DEVICE)

print "Iterations = %d" % (epochs)

start_time = timeit.default_timer()
# lo = model.train(images, labels)
lo = model.train_lbfgs(images, labels, epochs)
elapsed = timeit.default_timer() - start_time
print "Cuda training using L-BFGS took %f seconds" % (elapsed)

# Calculate test accuracy
outputs = model(test_images)
# predicted = torch.round(torch.sigmoid(outputs))
predicted = torch.round(outputs)
correct= (predicted == test_labels).sum()
accuracy = 100. * float(correct.item())/float(total_test)
print "L-BFGS: train loss = %f, test accuracy = %f" % (lo, accuracy)

########################################################
############################ CPU Training
########################################################
print "Training pytorch models using CPU"
DEVICE = torch.device("cpu")
########################################################
############################ Gradient Descent
########################################################
epochs = 3000
# Model and training set up
model = LogisticRegression(input_dim, output_dim).to(DEVICE)
criterion = torch.nn.BCELoss().to(DEVICE)
optimizer = torch.optim.SGD(model.parameters(), lr=lr_rate)

y_train.resize(len(y_train),1)
y_test.resize(len(y_test),1)
total = len(y_train)
total_test = len(y_test)
images = Variable(torch.from_numpy(X_train)).float()
labels = Variable(torch.from_numpy(y_train)).float()
test_images = Variable(torch.from_numpy(X_test)).float()
test_labels = Variable(torch.from_numpy(y_test)).float()

images = images.to(DEVICE)
labels = labels.to(DEVICE)
test_images = test_images.to(DEVICE)
test_labels = test_labels.to(DEVICE)

print "Iterations = %d" % (epochs)

start_time = timeit.default_timer()
lo = model.train(images, labels, epochs)
elapsed = timeit.default_timer() - start_time
print "CPU training using gradient descent took %f seconds" % (elapsed)

# Calculate test accuracy
outputs = model(test_images)
predicted = torch.round(outputs)
correct= (predicted == test_labels).sum()
accuracy = 100. * float(correct.item())/float(total_test)
print "Gradient descent: train loss = %f, test accuracy = %f" % (lo, accuracy)

########################################################
############################ L-BFGS
########################################################

epochs = 100
# Model and training set up
model = LogisticRegression(input_dim, output_dim).to(DEVICE)
criterion = torch.nn.BCELoss().to(DEVICE)
optimizer = torch.optim.LBFGS(model.parameters()) 

y_train.resize(len(y_train),1)
y_test.resize(len(y_test),1)
total = len(y_train)
total_test = len(y_test)
images = Variable(torch.from_numpy(X_train)).float()
labels = Variable(torch.from_numpy(y_train)).float()
test_images = Variable(torch.from_numpy(X_test)).float()
test_labels = Variable(torch.from_numpy(y_test)).float()

images = images.to(DEVICE)
labels = labels.to(DEVICE)
test_images = test_images.to(DEVICE)
test_labels = test_labels.to(DEVICE)

print "Iterations = %d" % (epochs)

start_time = timeit.default_timer()
# lo = model.train(images, labels)
lo = model.train_lbfgs(images, labels, epochs)
elapsed = timeit.default_timer() - start_time
print "CPU training using L-BFGS took %f seconds" % (elapsed)

# Calculate test accuracy
outputs = model(test_images)
# predicted = torch.round(torch.sigmoid(outputs))
predicted = torch.round(outputs)
correct= (predicted == test_labels).sum()
accuracy = 100. * float(correct.item())/float(total_test)
print "L-BFGS: train loss = %f, test accuracy = %f" % (lo, accuracy)
