#!/bin/bash
#######################################################
# 
# Written by John Halloran <jthalloran@ucdavis.edu>
#
# Copyright (C) 2020 John Halloran
# Licensed under the Open Software License version 3.0
# See COPYING or http://opensource.org/licenses/OSL-3.0
#
#######################################################
#
# If you use any of the code below, please cite:
# J. T. Halloran and D. M. Rocke.
# GPU-Accelerated Primal Learning for Extremely Fast Large-Scale Classification,
# Advances in Neural Information Processing Systems (NeurIPS) 2020
#
#######################################################


######################################### Description ######################################### 
############################################################################################### 
# 
# This bash script, run.sh, builds the various logistic regression implementations from the paper, 
# then downloads the benchmark datasets, 
# then runs the different implementations multiple times.
# CUDA 10.0 (or greater) must be installed and in your path, as well as g++/gcc and CMake.
# 
# The following are descriptions of the various functions in the bash script, which may be commented out if you've already
# either run the build command or downloaded the data:
# buildAll: build all the implementations on your machine
# dlData: download the datasets
# testAll: run the timing tests.  Note that the the variable NR should be set to the maximum number of supported threads on your machine
#
# Logistic regression implementations and short descriptions:
# tronLrMix: TRON-LR-MIX in the paper, the GPU+multithread-optimized version of TRON
# tronLrGpu0: TRON-LR-GPU0 in the paper, the straight-forward GPU implementation
# tronLrGpu: TRON-LR-GPU in the paper, the highly-optimized GPU implementation from the paper
# liblinear: standard single-threaded version of LIBLINEAR used to measure speedups in the paper
# liblinear-multicore-2.30: TRON-LR-CPU in the paper, the multithread-optimized version of LIBLINEAR

# Datasets:
# https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/rcv1_train.binary.bz2
# https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/SUSY.bz2
# https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/HIGGS.bz2
# https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/kddb.bz2
# https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/url_combined.bz2
# https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/real-sim.bz2
# 
############################################################################################### 
############################################################################################### 

function dlData {
    mkdir data
    cd data
    D0="https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/"
    for d in rcv1_train.binary.bz2 SUSY.bz2 HIGGS.bz2 kddb.bz2 url_combined.bz2 real-sim.bz2
    do
	echo $d
	wget "$D0/$d"
	bzip2 -d $d
    done
    cd ..
}

function makeGpu {
    D0=$(pwd)
    D=$1
    cd $D
    cmake .
    make
    cd $D0
}

function makeCpu {
    D0=$(pwd)
    D=$1
    cd $D
    make
    cd $D0
}

function buildAll {
    # Make GPU builds
    for b in tronLrGpu0 tronLrGpu tronLrMix
    do
	x="$(pwd)/$b"
	echo $x
	makeGpu $x
    done

    # Make CPU builds
    for b in liblinear liblinear-multicore-2.30
    do
	x="$(pwd)/$b"
	echo $x
	makeCpu $x
    done
}

function testAll {
    D0=$(pwd)
    for d in $(ls data/*)
    do
	# Test single-threaded solvers
	for lrSolver in liblinear tronLrGpu0 tronLrGpu
	do
	    echo "Testing $lrSolver, $d" | tee -a $F
	    solver="${D0}/${lrSolver}/train"
	    # output file
	    F="${lrSolver}.txt"
	    # run solver
	    $solver -c 4 -e 0.1 -s 0 $d | tee -a $F
	done

	# Test multithreaded implementations
	lrSolver=tronLrMix
	# output file
	F="${lrSolver}.txt"
	echo "Testing $lrSolver, $d" | tee -a $F
	solver="${D0}/${lrSolver}/train"
	# TRON-LR-MIX, GPU and multithread optimized solver
	for ((nr=1;nr<=${NR};nr++))
	do
	    # run solver
	    $solver -c 4 -nr $nr -e 0.1 -s 0 $d | tee -a $F
	done

	# TRON-LR-CPU, liblinear multicore implementation
	echo "Testing tronLrCpu, $d" | tee -a $F
	lrSolver=liblinear-multicore-2.30
	# output file
	F="tronLrCpu.txt"
	solver="${D0}/${lrSolver}/train"
	for ((nr=1;nr<=${NR};nr++))
	do
	    for ((i=0;i<10;i++))
	    do
		# run solver
		$solver -c 4 -nr $nr -e 0.1 -s 0 $d | tee -a $F
	    done
	done
    done
}

# Build all logistic regression solvers
buildAll

# Download the datasets
dlData

# run all the timing tests
NR=48
export OMP_NUM_THREADS=$NR
testAll
