#!/bin/bash
#######################################################
# 
# Written by John Halloran <jthalloran@ucdavis.edu>
#
# Copyright (C) 2020 John Halloran
# Licensed under the Open Software License version 3.0
# See COPYING or http://opensource.org/licenses/OSL-3.0
#
#######################################################
#
# If you use any of the code below, please cite:
# J. T. Halloran and D. M. Rocke.
# GPU-Accelerated Primal Learning for Extremely Fast Large-Scale Classification,
# Advances in Neural Information Processing Systems (NeurIPS) 2020
#
#######################################################

############################################################################################### 
######################################### Description ######################################### 
############################################################################################### 
# 
# This bash script, run.sh, builds the various logistic regression implementations from the paper, 
# then downloads the benchmark datasets, 
# then runs the different implementations multiple times.
# CUDA 10.0 (or greater) must be installed and in your path, as well as g++/gcc and CMake.
# 
# The following are descriptions of the various functions in the bash script, which may be commented out if you've already
# either run the build command or downloaded the data:
# buildAll: build all the implementations on your machine
# dlData: download the datasets
# testAll: run the timing tests.  Note that the the variable NR should be set to the maximum number of supported threads on your machine
#
# SVM proteomics classificatio solvers (implemented in Percolator) and short descriptions:
# tronSvmGpu: TRON-SVM-GPU in the paper, the highly-optimized GPU implementation from the paper
# tronSvmMix: TRON-SVM-MIX in the paper, the GPU+multithread-optimized version of TRON, trades off speed for memory compared to tronSvmGpu
# tronSvmCpu: TRON-SVM-CPU in the paper, the highly-optimized multithreaded version of TRON from the 2018 Halloran and Rocke paper:
#                  "A Matter of Time: Faster Percolator Analysis via Efficient SVM Learning for Large-Scale Proteomics"
#                  Note that L2-SVM-MFN (the highly-optimized multithreaded implementation described in the above paper) is also included in the tronSvmCpu implementation
# percolator: current standard implementation of Percolator, which uses a single-threaded version of L2-SVM-MFN (adapted from Sindhwani's 2006 SVMlin software)

# Large-scale datasets with freely available permissions to distribute:
# http://jthalloran.ucdavis.edu/wuDataset.pin.gz
# http://jthalloran.ucdavis.edu/kimDataset.pin.gz
# 
############################################################################################### 
############################################################################################### 

function dlData {
    mkdir data
    cd data
    # Get large-scale datasets with distribution permissions from the Halloran and Rocke 2018 paper,
    # "A Matter of Time: Faster Percolator Analysis via Efficient SVM Learning for Large-Scale Proteomics"
    D0="http://jthalloran.ucdavis.edu/"
    for d in wuDataset.pin.gz kimDataset.pin.gz
    do
	echo $d
	wget "$D0/$d"
	gunzip $d
    done
    cd ..
}

function makeIt {
    D0=$(pwd)
    D=$1
    cd $D
    cmake .
    make
    cd $D0
}

function buildAll {
    # Make GPU and CPU builds
    for b in percolator_tronSvmGpu percolator_tronSvmMix percolator_tronSvmCpu percolator
    do
	x="$(pwd)/$b"
	echo $x
	makeIt $x
    done
}

function testAll {
    D0=$(pwd)
    for d in $(ls data/*)
    do
	#############################################################
	# Test with baseline solver, Percolator's current SVM solver
	#############################################################
	svmSolver=percolator
	echo "Testing $svmSolver, $d" | tee -a $F
	perc="${D0}/${svmSolver}/src/percolator"
	# output file
	F="${svmSolver}.txt"
	# run solver
	$perc $d -U -M d.txt -r t.txt 2>&1 | tee -a $F

	#############################################################
	# Test GPU-optimized version of TRON in Percolator
	#############################################################
	svmSolver=percolator_tronSvmGpu
	echo "Testing $svmSolver, $d" | tee -a $F
	perc="${D0}/${svmSolver}/src/percolator"
	# output file
	F="${svmSolver}.txt"
	# run solver
	$perc $d -U -M d.txt -r t.txt -TRON -nr 1 2>&1 | tee -a $F

	#############################################################
	# Test multithreaded TRON implementations
	#############################################################
	for svmSolver in percolator_tronSvmMix percolator_tronSvmCpu
	do
	    # output file
	    F="${svmSolver}.txt"
	    echo "Testing $svmSolver, $d" | tee -a $F
	    perc="${D0}/${svmSolver}/src/percolator"
	    for ((nr=8;nr<=${NR};nr+=8))
	    do
		# run solver
		$perc $d -U -M d.txt -r t.txt -TRON -nr ${nr} 2>&1 | tee -a $F
	    done
	done

	#############################################################
	# Test multithreaded L2-SVM-MFN
	#############################################################
	svmSolver=percolator_tronSvmMix
	# output file
	F="${svmSolver}.txt"
	echo "Testing $svmSolver, $d" | tee -a $F
	perc="${D0}/${svmSolver}/src/percolator"
	for ((nr=8;nr<=${NR};nr+=8))
	do
	    # run solver
	    $perc $d -U -M d.txt -r t.txt -nr ${nr} 2>&1 | tee -a $F
	done

    done
}

# Build all logistic regression solvers
buildAll

# Download the datasets
dlData

# run all the timing tests
NR=48
export OMP_NUM_THREADS=$NR
testAll
